/*!\file mobile.c
 *
 * \brief Gestion, déplacement et affichage de mobiles à l'écran.
 *
 * \author Farès BELHADJ, amsi@ai.univ-paris8.fr
 * \date May 15 2014
 */

#include <primitives.h>
#include <assert.h>

typedef struct mobile_t mobile_t;

struct mobile_t {
  double px, py; /* position     */
  double vx, vy; /* vitesse      */
  double ax, ay; /* accélération */
  int r;         /* rayon        */
  Uint32 c;      /* couleur      */
};

static int _nb_mobiles = 0;
static mobile_t * _mobiles = NULL;

void mobileInit(int n) {
  int i;
  Uint8 r, g, b;
  assert(n > 0);
  if(_mobiles == NULL)
    _mobiles = malloc((_nb_mobiles = n) * sizeof *_mobiles);
  else if(n != _nb_mobiles)
    _mobiles = realloc(_mobiles, (_nb_mobiles = n) * sizeof *_mobiles);
  assert(_mobiles);
  for(i = 0; i < _nb_mobiles; i++) {
    r = 256 * (rand() / (RAND_MAX + 1.0));
    g = 256 * (rand() / (RAND_MAX + 1.0));
    b = 256 * (rand() / (RAND_MAX + 1.0));
    _mobiles[i].px = primGetScreenWidth()  >> 1;
    _mobiles[i].py = primGetScreenHeight() >> 1;
    _mobiles[i].vx = primGetScreenWidth()  * (1.0 - 2.0 * rand() / (RAND_MAX + 1.0));
    _mobiles[i].vy = primGetScreenHeight() * (-1.0 * rand() / (RAND_MAX + 1.0));
    _mobiles[i].ax = 0.0;
    _mobiles[i].ay = 900.8;
    _mobiles[i].r  = 30 * (rand() / (RAND_MAX + 1.0));
    _mobiles[i].c  = RGB(r, g, b);
  }
}

void mobileMove(void) {
  int i;
  double dt = 1.0 / 60.0, ff = 0.8;
  static Uint32 t0 = 0;
  Uint32 t = SDL_GetTicks();
  dt = (t - t0) / 1000.0;
  t0 = t;
  for(i = 0; i < _nb_mobiles; i++) {
    _mobiles[i].px += _mobiles[i].vx * dt;
    _mobiles[i].py += _mobiles[i].vy * dt;
    _mobiles[i].vx += _mobiles[i].ax * dt;
    _mobiles[i].vy += _mobiles[i].ay * dt;
    if((_mobiles[i].px + _mobiles[i].vx * dt - _mobiles[i].r < 0.0 && _mobiles[i].vx < 0) ||
       (_mobiles[i].px + _mobiles[i].vx * dt + _mobiles[i].r >= primGetScreenWidth() && _mobiles[i].vx > 0) ) {
      _mobiles[i].vx -= _mobiles[i].ax * dt;
      _mobiles[i].vx = -ff * _mobiles[i].vx;
      _mobiles[i].vy = ff * _mobiles[i].vy;
    }
    if((_mobiles[i].py + _mobiles[i].vy * dt - _mobiles[i].r < 0.0 && _mobiles[i].vy < 0.0) ||
       (_mobiles[i].py + _mobiles[i].vy * dt + _mobiles[i].r >= primGetScreenHeight() && _mobiles[i].vy > 0.0) ) {
      _mobiles[i].vy -= _mobiles[i].ay * dt;
      _mobiles[i].vx = ff * _mobiles[i].vx;
      _mobiles[i].vy = -ff * _mobiles[i].vy;
    }
  }
}

void mobileDraw(void) {
  int i;
  primClearScreen();
  for(i = 0; i < _nb_mobiles; i++) {
    primSetColor(_mobiles[i].c);
    primFilledCircle((int)_mobiles[i].px, (int)_mobiles[i].py, _mobiles[i].r);
  }
}

void mobileQuit(void) {
  if(_mobiles) {
    free(_mobiles);
    _mobiles = NULL;
  }
  _nb_mobiles = 0;
}

