/*!\file primitives.c
 *
 * \brief Primitives de dessin 2D sur texture GL.
 *
 * Nouvelle version, gère de multiple écrans.
 * \author Farès BELHADJ, amsi@ai.univ-paris8.fr
 * \date March 14 2014
 */
#include "primitives.h"
#include <miscHelper.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <assert.h>

typedef struct screen_node_t screen_node_t;
struct screen_node_t {
  Uint32 * pixels;
  GLuint w, h, tId;
  struct screen_node_t * next;
};

static screen_node_t * _screen_list = NULL, ** _cur_screen = &_screen_list;
static void addScreen(GLuint w, GLuint h);

/*!\brief identifiant du programme GLSL */
static GLuint _pId = 0; 

/*!\brief La couleur en cours ; a modifier avec les fonctions setColor */
static Uint32 _cur_color = 0xFF0000FF;

/*!\brief revoie la \a _cur_color. */
Uint32 primGetColor(void) {
  return _cur_color;
}

/*!\brief met dans \a _cur_color la couleur passee en argument. */
void primSetColor(Uint32 color) {
  _cur_color = color;
}

/*!\brief initialise (ou réinitialise) l'écran aux dimensions \a w et
 * \a h, le lie à l'identifiant (OpenGL) de texture retourné.
 * 
 * \param w la largeur de la fenêtre/texture GL.
 * \param h la hauteur de la fenêtre/texture GL.
 * \return l'identifiant (OpenGL) de la texture générée.
 */
GLuint primInitScreenWithDimensions(GLuint w, GLuint h) {
  assert(w * h);
  addScreen(w, h);
  glBindTexture(GL_TEXTURE_2D, (*_cur_screen)->tId);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
  if(_pId == 0)
    _pId = gl4duCreateProgram("<vs>shaders/basic.vs", "<fs>shaders/basic.fs", NULL);
  glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, w, h, 0, GL_RGBA, GL_UNSIGNED_BYTE, NULL);
  mhInit();
  return (*_cur_screen)->tId;
}

/*!\brief initialise (ou réinitialise) l'écran aux dimensions du
 * viewport OpenGL, le lie à l'identifiant (OpenGL) de texture
 * retourné.
 * 
 * Utilise \param initScreenWithDimensions.
 *
 * \return l'identifiant (OpenGL) de la texture générée.
 * \see initScreenWithDimensions
 */
GLuint primInitScreen(void) {
  int vp[4];
  glGetIntegerv(GL_VIEWPORT, vp);
  return primInitScreenWithDimensions(vp[2], vp[3]);
}

/*!\brief ajoute un écran à la tête de liste d'écrans gérés.
 * 
 * \param w la largeur de l'écran à créer.
 * \param h la hauteur de l'écran à créer.
 */
static void addScreen(GLuint w, GLuint h) {
  screen_node_t * newscr = NULL;
  assert(w * h);
  newscr = malloc(sizeof *newscr);
  assert(newscr);
  glGenTextures(1, &(newscr->tId));
  assert(newscr->tId);
  newscr->pixels = calloc(w * h, sizeof *(newscr->pixels));
  assert(newscr->pixels);
  newscr->w = w;
  newscr->h = h;
  newscr->next = _screen_list;
  _screen_list = newscr;
  _cur_screen = &_screen_list;
}

/*!\brief active comme écran courant l'écran dont l'id est passé en
 * argument.
 * 
 * \param id identifiant de l'écran.
 */
int primSetScreen(GLuint id) {
  screen_node_t ** c = &_screen_list;
  while(*c && (*c)->tId != id)
    c = &((*c)->next);
  if(*c) {
    _cur_screen = c;
    return 1;
  }
  return 0;
}

/*!\brief supprime l'écran et se repositionne sur le premier */
void primDeleteScreen(void) {
  screen_node_t * to_delete;
  mhQuit();
  if(!*_cur_screen) return;
  if((*_cur_screen)->pixels != NULL)
    free((*_cur_screen)->pixels);
  if((*_cur_screen)->tId)
    glDeleteTextures(1, &((*_cur_screen)->tId));
  to_delete = *_cur_screen;
  *_cur_screen = (*_cur_screen)->next;
  free(to_delete);
  _cur_screen = &_screen_list;
}

/*!\brief retourne l'identifiant (OpenGL) de la texture liée à l'écran en cours.
 * \return l'identifiant (OpenGL) de la texture liée à cette bibliothèque.
 */
GLuint primGetTextureId(void) {
  return (*_cur_screen)->tId;
}

/*!\brief retourne la largeur de l'écran. */
GLuint primGetScreenWidth(void) {
  return (*_cur_screen)->w;
}

/*!\brief retourne la hauteur de l'écran. */
GLuint primGetScreenHeight(void) {
  return (*_cur_screen)->h;
}

/*!\brief Efface l'écran en mettant 0. */
void primClearScreen(void) {
  memset((*_cur_screen)->pixels, 0, (*_cur_screen)->w * (*_cur_screen)->h * sizeof *(*_cur_screen)->pixels);
}

/*!\brief Efface l'écran en y mettant la valeur \a color. */
void primClearScreenWith(Uint32 color) {
  GLuint wh = (*_cur_screen)->w * (*_cur_screen)->h, i;
  for(i = 0; i < wh; i++)
    (*_cur_screen)->pixels[i] = color;
}

/*!\brief renvoie la couleur à la coordonnée (x, y) */
Uint32 primGetPixel(int x, int y) {
  return (*_cur_screen)->pixels[y * (*_cur_screen)->w + x];
}

/*!\brief met la couleur en cours à la coordonnée (x, y) */
void primPutPixel(int x, int y) {
  (*_cur_screen)->pixels[y * (*_cur_screen)->w + x] = _cur_color;
}

/*!\brief met à jour l'écran en envoyant la sous texture de dimensions
 * \a rect à GL. Si rect vaut NULL, met à jour toute la texture.
 *
 * \param rect le pointeur vers les quatre entiers positifs x,y (coin haut gauche du rectangle) et w,h (les dimensions du rectangle).
 */
void primUpdateScreen(GLint * rect) {
  glBindTexture(GL_TEXTURE_2D, (*_cur_screen)->tId);
  if(rect == NULL)
    glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, (*_cur_screen)->w, (*_cur_screen)->h, 0, GL_RGBA, GL_UNSIGNED_BYTE, (*_cur_screen)->pixels);
  else {
    glPixelStorei(GL_UNPACK_ROW_LENGTH, (*_cur_screen)->w);
    glTexSubImage2D(GL_TEXTURE_2D, 0, rect[0], rect[1], rect[2], rect[3], GL_RGBA, GL_UNSIGNED_BYTE, &(*_cur_screen)->pixels[rect[0] + rect[1] * (*_cur_screen)->w]);
    glPixelStorei(GL_UNPACK_ROW_LENGTH, 0);
  }
  glDisable(GL_DEPTH);
  glUseProgram(_pId);
  glBindVertexArray(mhGetMappedQuadVAO());
  glEnableVertexAttribArray(0);
  glEnableVertexAttribArray(1); 
  glActiveTexture(GL_TEXTURE0);
  glBindTexture(GL_TEXTURE_2D, (*_cur_screen)->tId);
  glUniform1i(glGetUniformLocation(_pId, "myTexture"), 0);
  glDrawArrays(GL_TRIANGLE_STRIP, 0, 4);
  glDisableVertexAttribArray(1);
  glDisableVertexAttribArray(0);
}

/*!\brief dessine un rectangle
 * \a rect en utilisant la couleur en cours.
 *
 * \param rect le pointeur vers les quatre entiers positifs x,y (coin haut gauche du rectangle) et w,h (les dimensions du rectangle).
 */
void primRect(GLint * rect) {
  GLint x, y, yw, mx = rect[0] + rect[2], my = rect[1] + rect[3];
  for(y = rect[1]; y < my; y++) {
    yw = y * (*_cur_screen)->w;
    for(x = rect[0]; x < mx; x++)
      (*_cur_screen)->pixels[x + yw] = _cur_color;
  }
}

/*!\brief dessine un segment (x0, y0) -> (x1, y1) de couleur \a
 * _cur_color en utilisant l'algorithme de Bresenham'65.
 */
void primLine(int x0, int y0, int x1, int y1) {
  int u = x1 - x0, v = y1 - y0;
  int pasX = (u < 0) ? -1 : 1, pasY = (v < 0) ? -1 : 1;
  int x, y, del, incH, incO;
  if(abs(u) < abs(v)) { /* deuxieme octant */
    del = (incH = ((pasX * u) << 1)) - pasY * v;
    incO = incH - ((pasY * v) << 1);
    for(y = y0, x = x0; y != y1; y += pasY) {
      if(IN_SCREEN(x, y))
	primPutPixel(x, y);
      if(del < 0) del += incH;
      else {
	del += incO;
	x += pasX;
      }
    }
  } else {  /* premier octant */
    del = (incH = ((pasY * v) << 1)) - pasX * u;
    incO = incH - ((pasX * u) << 1);
    for(x = x0, y = y0; x != x1; x += pasX) {
      if(IN_SCREEN(x, y))
	primPutPixel(x, y); 
      if(del < 0) del += incH;
      else {
	del += incO;
	y += pasY;
      }
    }
  }
}

/*!\brief dessine un cercle centré en (x0, y0) de rayon r en
 * utilisant l'algorithme de Bresenham'77.
 */
void primCircle(int x0, int y0, int r) {
  int x, y, del, incH, incO, t = M_SQRT1_2 * r + 1;
  del = 3 - (r << 1);
  incH = 6;
  incO = 10 - (r << 2);
  for(x = 0, y = r; x <= t; x++, incH += 4, incO += 4) {
    if(IN_SCREEN(x0 + x, y0 + y)) primPutPixel(x0 + x, y0 + y);
    if(IN_SCREEN(x0 + x, y0 - y)) primPutPixel(x0 + x, y0 - y);
    if(IN_SCREEN(x0 - x, y0 + y)) primPutPixel(x0 - x, y0 + y);
    if(IN_SCREEN(x0 - x, y0 - y)) primPutPixel(x0 - x, y0 - y);
    if(IN_SCREEN(x0 + y, y0 + x)) primPutPixel(x0 + y, y0 + x);
    if(IN_SCREEN(x0 + y, y0 - x)) primPutPixel(x0 + y, y0 - x);
    if(IN_SCREEN(x0 - y, y0 + x)) primPutPixel(x0 - y, y0 + x);
    if(IN_SCREEN(x0 - y, y0 - x)) primPutPixel(x0 - y, y0 - x);
    if(del < 0) del += incH;
    else {
      y--;
      incO += 4;
      del += incO;
    }
  }
}

/*!\brief dessine un segment horizontal (x0, y) -> (x1, y).
 */
void primHLine(int x0, int x1, int y) {
  int x, pasX = (x1 - x0) < 0 ? -1 : 1, x1pp;
  y  = MIN(MAX(0, y), primGetScreenHeight() - 1);
  x0 = MIN(MAX(0, x0), primGetScreenWidth() - 1);
  x1 = MIN(MAX(0, x1), primGetScreenWidth() - 1);
  x1pp = x1 + pasX;
  for(x = x0; x != x1pp; x += pasX)
    primPutPixel(x, y);
}

/*!\brief dessine un cercle plein centré en (x0, y0) de rayon r
 * en utilisant l'algorithme de Bresenham'77.
 *
 * \see hLine
 */
void primFilledCircle(int x0, int y0, int r) {
  int x, y, del, incH, incO, t = M_SQRT1_2 * r + 1;
  del = 3 - (r << 1);
  incH = 6;
  incO = 10 - (r << 2);
  for(x = 0, y = r; x <= t; x++, incH += 4, incO += 4) {
    primHLine(x0 + x, x0 - x, y0 + y);
    primHLine(x0 + x, x0 - x, y0 - y);
    primHLine(x0 + y, x0 - y, y0 + x);
    primHLine(x0 + y, x0 - y, y0 - x);
    if(del < 0) del += incH;
    else {
      y--;
      incO += 4;
      del += incO;
    }
  }
}
